# Copyright (c) Facebook, Inc. and its affiliates. (http://www.facebook.com)
# -*- coding: utf-8 -*-

"""
mapillary.config.api.__init__

This package contains all the API v4 endpoints provided with the Mapillary Python SDK.

- Copyright: (c) 2021 Facebook
- License: MIT LICENSE
"""

from . import vector_tiles  # noqa: F401
from . import entities  # noqa: F401
from . import general  # noqa: F401
